
from astropy import units as u

from gala import potential as gp
from gala.units import galactic



def JZ2023_1e12():
    """ 10^12 Msun NFW host halo from Zhu+2023
        (https://iopscience.iop.org/article/10.3847/1538-4357/acfe6f)

    Returns:
        gala potential: Gala NFW potential
    """
    c_NFW = 10**0.945
    M_host = 1e12 * u.Msun
 
    return  gp.NFWPotential.from_M200_c(M200=M_host, 
                                        c=c_NFW, 
                                        units=galactic)


def JZ2023_1e13():
    """ 10^13 Msun NFW host halo from Zhu+2023
        (https://iopscience.iop.org/article/10.3847/1538-4357/acfe6f)

    Returns:
        gala potential: Gala NFW potential
    """
    c_NFW = 10**0.85
    M_host = 1e13 * u.Msun
 
    return  gp.NFWPotential.from_M200_c(M200=M_host, 
                                        c=c_NFW, 
                                        units=galactic)

def JZ2023_1e14():
    """ 10^14 Msun NFW host halo from Zhu+2023
        (https://iopscience.iop.org/article/10.3847/1538-4357/acfe6f)
    
    Returns:
        gala potential: Gala NFW potential
    """
    c_NFW = 10**0.75
    M_host = 1e14 * u.Msun
 
    return  gp.NFWPotential.from_M200_c(M200=M_host, 
                                        c=c_NFW, 
                                        units=galactic)



from astropy import units as u
from gala.units import galactic
import numpy as np

import galarp as grp


__all__ = ['ExampleUniformEvent', 'ExampleLorentzianEvent']


def ExampleUniformEvent(inc=45., strength=600.):
    """ Returns a set of orbits for a simple uniform wind event.

    Args:
        inc (float, optional): Inclination of the wind vector (in deg). Defaults to 45.
        strength (float, optional): Strength of the wind vector (in km/s). Defaults to 600.

    Returns:
        galarp OrbitContainer: The orbit output.
    """
    pot = grp.builtins.JZ2023_Satellite()
    mass_profile = grp.gen_mass_profile(pot)
    
    wind = grp.RPWind(units=galactic)
    wind.init_from_inc(inclination=np.deg2rad(inc), strength= strength *u.km/u.s)
    
    shadow = grp.UniformShadow()
    shadow.init_from_wind(wind)

    particles = grp.UniformGrid(Rmax= 10, n_particles= 20, z_start= 0.5)
    particles.generate(mass_profile)

    sim = grp.RPSim(wind=wind, potential=pot, shadow=shadow, rho_icm=1e-26 * u.g/ u.cm**3)
    orbits = sim.run(particles, rho_icm=1e-26 * u.g/ u.cm**3, integration_time=1000 * u.Myr, dt=2*u.Myr)
    return orbits


def ExampleLorentzianEvent(inc=45, strength=600):
    """ Returns a set of orbits for a simple Lorentzian wind event.

    Args:
        inc (float, optional): Inclination of the wind vector (in deg). Defaults to 45.
        strength (float, optional): Strength of the wind vector (in km/s). Defaults to 600.

    Returns:
        galarp OrbitContainer: The orbit output.
    """

    pot = grp.builtins.JZ2023_Satellite()
    mass_profile = grp.gen_mass_profile(pot)
    wind = grp.LorentzianWind(t0=500 * u.Myr, width = 500 * u.Myr, units=galactic)
    wind.init_from_inc(inclination=np.deg2rad(inc), strength= strength *u.km/u.s)
    
    shadow = grp.UniformShadow()
    shadow.init_from_wind(wind)

    particles = grp.UniformGrid(Rmax= 10, n_particles= 20, z_start= 0.5)
    particles.generate(mass_profile)

    sim = grp.RPSim(wind=wind, potential=pot, shadow=shadow, rho_icm=1e-26 * u.g/ u.cm**3)
    orbits = sim.run(particles, rho_icm=1e-26 * u.g/ u.cm**3, integration_time=1000 * u.Myr, dt=2*u.Myr)
    return orbits


# Plotting routines and convenience functions.

# 3rd Party
from astropy.cosmology import WMAP9 as cosmo
from astropy import constants as c
from astropy import units as u
import numpy as np


from scipy.interpolate import interp1d

import pickle


__all__ = ['v_circ', 'velocity', 'gen_mass_profile', 'm0', 'get_orbit_data', 'pickle_obj', 'Rvir', 'R100', 'R500', 'rotate']


#############################################
######## Convenience Functions ##############
#############################################


def v_circ(M, R):
    return np.sqrt(c.G * M / R).to(u.km / u.s)


def velocity(mass_profile, R, theta):
    vx = -v_circ(mass_profile(R) * u.M_sun, R) * np.sin(theta)
    vy = v_circ(mass_profile(R) * u.M_sun, R) * np.cos(theta)
    return [vx.value, vy.value, 0] * (u.km / u.s)


def gen_mass_profile(potential, lims=[2e-2, 80]):
    pos = np.zeros((3, 100)) * u.kpc
    pos[0] = np.linspace(lims[0], lims[1], 100) * u.kpc
    m_profile = potential.mass_enclosed(pos)
    return interp1d(pos[0], m_profile, bounds_error=False, fill_value=0)


def m0(rho_d0):
    """Return the normalization constant for the Burkert Dark Matterpotential

    Args:
        rho_d0 (float): Central density of the Burkert potential (in g/cm^3)
    """
    return 1e9 * (rho_d0 / 1.46e-24) ** (-7 / 2)


def get_orbit_data(o):
    pos, vel = o.pos, o.vel

    x, y, z = pos.xyz.value
    x, y, z = x.T, y.T, z.T

    vx, vy, vz = vel.d_xyz.to(u.km / u.s).value
    vx, vy, vz = vx.T, vy.T, vz.T

    return x, y, z, vx, vy, vz


def pickle_obj(obj, name="obj.out"):
    with open(name, "wb") as f:
        pickle.dump(obj, f)


def Rvir(mass):
    """ Returns the virial radius of the potential """
    
    mass = mass.to(u.Msun)

    rho_c = 3 * cosmo.H(0.)**2 / (8*np.pi*c.G)
    rho_c = rho_c.to(u.Msun/u.kpc**3)

    return np.cbrt(mass / (200*rho_c.value) / (4/3*np.pi))

def R100(mass):
    """ Returns the radius at which the density is 100 times the critical density """
    
    mass = mass.to(u.Msun)

    rho_c = 3 * cosmo.H(0.)**2 / (8*np.pi*c.G)
    rho_c = rho_c.to(u.Msun/u.kpc**3)

    return np.cbrt(mass / (100*rho_c.value) / (4/3*np.pi))

def R500(mass):
    """ Returns the radius at which the density is 500 times the critical density """
    
    mass = mass.to(u.Msun)

    rho_c = 3 * cosmo.H(0.)**2 / (8*np.pi*c.G)
    rho_c = rho_c.to(u.Msun/u.kpc**3)

    return np.cbrt(mass / (500*rho_c.value) / (4/3*np.pi))


def rotate(vec, alpha=np.deg2rad(0), beta=np.deg2rad(0), gamma=np.deg2rad(0)):
    """ Rotate a set of positions by the given angles

    Args:
        vec (array_like): Array of positions to rotate 
        alpha (float): Rotation angle about the x-axis
        beta (float): Rotation angle about the y-axis
        gamma (float): Rotation angle about the z-axis 

    Returns:
        _type_: _description_
    """

    rot_matrix_1 = [np.cos(beta)*np.cos(gamma), np.sin(alpha)*np.sin(beta)*np.cos(gamma) - np.cos(alpha)*np.sin(gamma), np.cos(alpha)*np.sin(beta)*np.cos(gamma) + np.sin(alpha)*np.sin(gamma)]
    rot_matrix_2 = [np.cos(beta)*np.sin(gamma), np.sin(alpha)*np.sin(beta)*np.sin(gamma) + np.cos(alpha)*np.cos(gamma), np.cos(alpha)*np.sin(beta)*np.sin(gamma) - np.sin(alpha)*np.cos(gamma)]
    rot_matrix_3 = [-np.sin(beta), np.sin(alpha)*np.cos(beta), np.cos(alpha)*np.cos(beta)]
    rot_matrix = np.array([rot_matrix_1, rot_matrix_2, rot_matrix_3])
    return np.dot(rot_matrix, vec)


from .densities import *            # Density classes
from .particle_grids import *       # ParticleGrid class
from .rampressure import *          # RPS classes and functions
from .shadows import *              # Shadow classes and functions
from .tests import *                # Testing utilities
from .utils import *                # Utility and plotting functions
from .winds import *                # Classes for ram pressure wind

from . import builtins

from .hosts import *
from .postprocessing import *

__version__ = "0.0.5"



import numpy as np
from astropy import units as u
from astropy import constants as const
from gala.units import UnitSystem

from gala import dynamics as gd
from gala import potential as gp

from matplotlib import pyplot as plt



class HostOrbit:
    """ Class for a host orbit to determine a ram pressure profile.
    """
    def __init__(self, potential, init_conditions=None, 
                 density=None, temp=2.51e6 * u.K):
        self.potential = potential

        self.init_conditions = init_conditions
        self.density = density
        self.temp = temp

        self.orbit = None
        self.velocities = None

    

    def integrate(self, n_steps=10000, dt=1):
        assert self.init_conditions is not None

        w0 = gd.PhaseSpacePosition(pos=self.init_conditions.pos, vel=self.init_conditions.vel)

        self.orbit = gp.Hamiltonian(self.potential).integrate_orbit(w0, dt=dt, n_steps=n_steps)
        

    def velocities(self, units = u.km/u.s):
        dx, dy, dz = self.orbit.vel
        v = np.sqrt(dx**2 + dy**2 + dz**2).to(units)
        self.velocities = v

    def gen_ENZO_filename(self, units = UnitSystem(u.cm, u.s, u.radian, u.g)
):
        
        dx, dy, dz = self.orbit.vel.d_xyz


    def plot_infall_orbit(self, outname=None):
        fig, ax = plt.subplots(2, 3, figsize=(13,8))
        x,y,z = self.orbit.pos.xyz
        r = np.sqrt(x**2 + y**2 + z**2)
        
        t = self.orbit.t

        dx, dy, dz = self.orbit.vel.d_xyz

        vel = np.sqrt(dx**2 + dy**2 + dz**2).to(u.km/u.s)
        color, lw = "black", 2

        ax[0][0].plot(x,y, color=color, lw=lw)
        ax[0][0].scatter(0, 0, marker="+", color="black", s=400)
        ax[0][0].set_ylabel('y [kpc]')
        ax[0][0].set_xticks([])

        ax[1][0].plot(x,z, color=color, lw=lw)
        ax[1][0].scatter(0, 0, marker="+", color="black", s=400)
        ax[1][0].set_xlabel('x [kpc]')
        ax[1][0].set_ylabel('z [kpc]')

        ax[0][1].plot(t, r, color=color, lw=lw)
        ax[0][1].set_ylabel('r [kpc]')
        ax[0][1].set_xticks([])

        ax[0][2].plot(t, vel, color=color, lw=lw)
        ax[0][2].set_ylabel('v [km/s]')
        ax[0][2].set_xticks([])

        if self.density is not None:
            dens = self.density.evaluate(r) * const.u.cgs
            ax[1][1].plot(t, dens, color=color, lw=lw)
            ax[1][1].set_ylabel('Density [g/cm^3]')

            rp_profile = (dens * vel**2).to(u.g / u.cm / u.s**2)
            

            ax[1][2].plot(t, rp_profile, color=color, lw=lw)
            ax[1][2].set_ylabel('Ram Pressure [g/cm/s^2]')

        ax[1][1].set_xlabel('t')
        ax[1][1].set_ylabel('Density [1e-3 cm^-3]')

        labels = ["XY", "Radius", "Total Velocity", "XZ", "Density", "RP Profile"]

        for i, axis in enumerate(ax.flatten()):
            xmin, xmax = axis.get_xlim()
            ymin, ymax = axis.get_ylim()
            dx, dy = xmax - xmin, ymax - ymin

            axis.text(xmax  - 0.3 * dx, ymin + 0.9 * dy, 
                    labels[i], fontsize=12, color='black', weight='bold')


        plt.tight_layout()
        if outname is not None:
            plt.savefig(outname)
        else:
            plt.show()



from .. import hosts
from .. import satellites

import gala.dynamics as gd
import gala.potential as gp


class BuiltInPotentialTestBase:

    name = None
    potential = None
    show_plots = False

    pos = [1, 0, 0]
    vel = [10, 10, 0]

    def test_creation(self):
        assert self.potential is not None
    
    def test_potential_integration(self, nsteps=1000):
        w0 = gd.PhaseSpacePosition(pos=self.pos, vel=self.vel)
        orbit = gp.Hamiltonian(potential=self.potential).integrate_orbit(w0, dt=1, n_steps=nsteps)
        
        assert len(orbit.pos) == nsteps + 1
        assert type(orbit) == gd.Orbit
    
    # def test_rs(self):
    #     mass = self.potential.mass_enclosed([500, 0., 0.] * u.Mpc)
        


class TestJZ2023_1e12(BuiltInPotentialTestBase):

    name = "JZ2023_1e12"
    potential = hosts.JZ2023_1e12()


class TestJZ2023_1e13(BuiltInPotentialTestBase):
    
        name = "JZ2023_1e13"
        potential = hosts.JZ2023_1e13()


class TestJZ2023_1e14(BuiltInPotentialTestBase):
         
        name = "JZ2023_1e14"
        potential = hosts.JZ2023_1e14()


class TestJZ2023_Satellite(BuiltInPotentialTestBase):
    
    name = "JZ2023_Satellite"
    potential = satellites.JZ2023_Satellite()


class TestNA2023_Satellite(BuiltInPotentialTestBase):
    name = "NA2023_Satellite"
    potential = satellites.NA2023_Satellite()




from ...utils import get_orbit_data

import numpy as np
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


class MatrixPlot:
    """ Base class to create matrix plots. """
    def __init__(self, **kwargs):
        self.cmap = kwargs.get("cmap", "viridis")
        self.figsize = kwargs.get("figsize", (10, 10))
        self.nrows = kwargs.get("nrows", 4)
        self.ncols = kwargs.get("ncols", 4)

        self.sharex = kwargs.get("sharex", True)
        self.sharey = kwargs.get("sharey", True)

    def create(self):
        self.fig, self.ax = plt.subplots(nrows=self.nrows, ncols=self.ncols, figsize=self.figsize, 
                               sharex=self.sharex, sharey=self.sharey)
        return self.fig, self.ax
    
    def populate(self, method, *args):
        index = 0
        for i in range(self.nrows):
            for j in range(self.ncols):
                method(i, j, index, *args)
                index += 1
    
    def add_xy_labels(self, xlabel, ylabel):
        for i in range(self.nrows):
            self.ax[i, 0].set_ylabel(ylabel)
        for j in range(self.ncols):
            self.ax[self.nrows - 1, j].set_xlabel(xlabel)
    
    def add_colorbar(self,i,j, mappable, loc="lower left"):
        ia = inset_axes(self.ax[i,j], width="50%", height="5%", loc=loc)
        plt.colorbar(mappable=mappable, cax=ia, orientation="horizontal")
        ia.xaxis.set_ticks_position("top")
    


def density_matrix(orbits, x_ind=0, y_ind=1, nrows=4, ncols=4, outname=None, **kwargs):
    """ Create a matrix plot of the density of the orbits for a given principle axis.

    Args:
        orbits (OrbitContainer): The orbits to plot.
        x_ind (int, optional): The x-axis index. Defaults to 0.
        y_ind (int, optional): The y-axis index. Defaults to 1.
        nrows (int, optional): The number of rows in the matrix. Defaults to 4.
        ncols (int, optional): The number of columns in the matrix. Defaults to 4.
        outname (str, optional): The name of the output file. Defaults to None.
    
    Usage:
        ```
        density_matrix(orbits, x_ind=0, y_ind=1, nrows=4, ncols=4, outname="test_out.pdf", **kwargs)
        ```
    """
    
    times = kwargs.get("times", np.linspace(0, len(orbits.data.t) - 1, nrows * ncols).astype(int))
    xextent = kwargs.get("xextent", (-40., 40.))
    xextent = (-xextent, xextent) if isinstance(xextent, (int, float)) else xextent
    yextent = kwargs.get("yextent", (-40., 40.))
    yextent = (-yextent, yextent) if isinstance(yextent, (int, float)) else yextent

    hlines, vlines = kwargs.get("hlines", []), kwargs.get("vlines", [])

    cbar_loc = kwargs.get("cbar_loc", "lower left")

    vmin, vmax = kwargs.get("vmin", 1), kwargs.get("vmax", 500)

    gridsize = kwargs.get("gridsize", 30)

    xlabel=kwargs.get("xlabel", "X (kpc)")
    ylabel=kwargs.get("ylabel", "Y (kpc)")
    
    mplot = MatrixPlot(**kwargs)
    fig, ax = mplot.create()

    data = get_orbit_data(orbits.data)

    x, y = data[x_ind], data[y_ind]

    mappables = []
    index = 0
    def method(i, j, index, x, y):
        this_x, this_y = x.T[times[index]], y.T[times[index]]

        hb = ax[i, j].hexbin(this_x, this_y, bins="log", cmap=mplot.cmap, gridsize=(gridsize, gridsize),
                        extent=[xextent[0], xextent[1], yextent[0], yextent[1]], 
                        vmin=vmin, vmax=vmax, zorder = 5)
        mappables.append(hb)
        xmin, xmax = ax[i, j].get_xlim()
        ymin, ymax = ax[i, j].get_ylim()
        dx, dy = xmax - xmin, ymax - ymin
        
        ax[i, j].text(xmin + dx/20, ymax - dy/10, f'{orbits.data.t[times[index]]}')
        for hline in hlines:
            ax[i, j].axhline(hline, color="Grey", lw=1, zorder=1)
        for vline in vlines:
            ax[i, j].axvline(vline, color="Grey", lw=1, zorder=1)

    mplot.populate(method, x, y)
    mplot.add_xy_labels(xlabel, ylabel)

    mplot.add_colorbar(i=nrows-1, j=ncols-1, mappable=mappables[-1], loc=cbar_loc)

    plt.subplots_adjust(hspace=0, wspace=0, left=0.07, right=0.99, top=0.99, bottom=0.05)

    if outname is not None:
        plt.savefig(outname)
    else:
        plt.show()



from astropy import units as u
from gala import potential as gp

from gala.units import galactic


def JZ2023_Satellite():
    """ Satellite potential for Zhu_+2023
        See https://iopscience.iop.org/article/10.3847/1538-4357/acfe6f
    """

    dm = gp.BurkertPotential(rho=5.93e-25 * u.g / u.cm**3, r0=11.87 * u.kpc, units=galactic)

    stars = gp.MiyamotoNagaiPotential(m=10**9.7 * u.Msun, a=2.5 * u.kpc, b=0.5 * u.kpc, units=galactic)
    gas = gp.MiyamotoNagaiPotential(m=10**9.7 * u.Msun, a=3.75 * u.kpc, b=0.75 * u.kpc, units=galactic)

    return gp.CompositePotential(dm=dm, stars=stars, gas=gas)


def NA2023_Satellite():
    """ Satellite potential for Akerman+2023 and Akerman+2024
        See https://iopscience.iop.org/article/10.3847/1538-4357/acbf4d
    """
    dm = gp.BurkertPotential.from_r0(r0=17.36 * u.kpc, units=galactic)
    stars = gp.MiyamotoNagaiPotential(1e11 * u.Msun, a= 5.94 * u.kpc, b = 0.58 * u.kpc, units=galactic)
    gas = gp.MiyamotoNagaiPotential(1e10 * u.Msun, a= 10.1 * u.kpc, b = 0.87 * u.kpc, units=galactic)

    return gp.CompositePotential(dm=dm, stars=stars, gas=gas)


from astropy import units as u
from ..hosts.initconditions import InitConditions


def JZ2023_1e12_IC():
    V200 = 143 
    V_phi = 0.655 * V200
    V_r = 0.832 * V200

    pos = [211, 0, 0] * u.kpc
    vel = [-V_r, V_phi, 0] * u.km / u.s

    return InitConditions(pos=pos, vel=vel)


def JZ2023_1e13_IC():
    V200 = 308 
    V_phi = 0.603 * V200
    V_r = 0.786 * V200

    pos = [455, 0, 0] * u.kpc
    vel = [-V_r, V_phi, 0] * u.km / u.s

    return InitConditions(pos=pos, vel=vel)


def JZ2023_1e14_IC():
    V200 = 663 
    V_phi = 0.53 * V200
    V_r = 0.782 * V200

    pos = [949, 0, 0] * u.kpc
    vel = [-V_r, V_phi, 0] * u.km / u.s

    return InitConditions(pos=pos, vel=vel)


from astropy import units as u
import numpy as np
import os

from .. import utils            # TODO This is horrible, but it works for now. Refactor the top-level utils file.

from matplotlib import pyplot as plt


__all__ = ["get_orbit_data", "k3d_plot", "plot_orbits", "plot_density", "plot_density_3ax"]

def get_orbit_data(o):
    pos, vel = o.pos, o.vel

    x, y, z = pos.xyz.value
    x, y, z = x.T, y.T, z.T

    vx, vy, vz = vel.d_xyz.to(u.km / u.s).value
    vx, vy, vz = vx.T, vy.T, vz.T

    return x, y, z, vx, vy, vz


def k3d_plot(
    orbit_containers,
    bgcolor=0,
    particle_color=0xFFFFFF,
    outname="test_out/orbits.html",
    duration=20,
    transpose=False,
    size=0.1,
    alpha=0.5,
):
    import k3d

    colors = [0xFFFFFF, 0x3387FF, 0xFF3333]

    wind = orbit_containers[0].metadata["WIND"].vector.to(u.km / u.s).value
    wind = np.array([wind[1], wind[0], wind[2]])
    wind /= np.sqrt(np.sum(wind**2))
    wind_length = 2
    wind_x0, wind_y0, wind_z0 = 0, 0, 0

    ell_xs, ell_ys, ell_zs = utils.ellipse_coords(0, 0, 10, 10, 0)

    plot = k3d.plot(
        fps=60,
        axes_helper=0,
        grid_visible=False,
        background_color=0,
    )

    for i, container in enumerate(orbit_containers):
        orbits = container.data

        pos = orbits.pos

        p_x, p_y, p_z = pos.xyz.value
        if transpose:
            p_x, p_y, p_z = p_x.T, p_y.T, p_z.T

        #vmin, vmax = -20, 20

        # v_z[v_z > vmax] = vmax
        # v_z[v_z < vmin] = vmin

        particles = k3d.points(
            np.vstack([p_x[0], p_y[0], p_z[0]]),
            point_size=size,
            color=colors[i],
            opacity=alpha,
        )
        plot += particles

        n_points = p_x.shape[0]

        t_int = np.arange(0, n_points, 1)
        t_sub = np.linspace(0, duration, n_points)

        particles.positions = {
            str(t): np.vstack([p_x[t_int[i]], p_y[t_int[i]], p_z[t_int[i]]]).T
            for i, t in enumerate(t_sub)
        }

    plot += k3d.line(np.vstack([ell_xs, ell_ys, ell_zs]).T, color=0xFFFFFF)

    plot += k3d.line(
        (
            [
                wind_x0,
                wind_x0 + wind[0] * wind_length,
                wind_y0,
                wind_y0 + wind[1] * wind_length,
                wind_z0,
                wind_z0 + wind[2] * wind_length,
            ]
        ),
        color=particle_color,
    )
    plot.display()

    with open(outname, "w") as fp:
        fp.write(plot.get_snapshot())


def plot_orbits(
    data,
    wind=None,
    shadow=None,
    plot_dir="plots/",
    R_plot=15,
    zrange=(-5, 15),
    title=None,
    plot_title="orbits",
):
    os.makedirs(plot_dir, exist_ok=True)

    fig, ax = plt.subplots(1, 3, figsize=(12, 5))

    pos = data.pos

    x, y, z = pos.xyz
    x, y, z = x.T, y.T, z.T

    for i in range(len(x)):
        ax[0].plot(x[i], y[i], color="Grey", alpha=0.3)
        ax[1].plot(x[i], z[i], color="Grey", alpha=0.3)
        ax[2].plot(y[i], z[i], color="Grey", alpha=0.3)

    ax[0].set_xlim(-R_plot, R_plot)
    ax[0].set_ylim(-R_plot, R_plot)
    ax[1].set_xlim(-R_plot, R_plot)
    ax[1].set_ylim(zrange[0], zrange[1])
    ax[2].set_xlim(-R_plot, R_plot)
    ax[2].set_ylim(zrange[0], zrange[1])

    utils.plot_disk(ax, 10)
    if wind is not None:
        utils.plot_wind_vector(
            wind.vector.value,
            ax,
            length=1,
            loc=(-R_plot + 1, -R_plot + 1, zrange[0] + 1),
            color="black",
        )

    if shadow is not None:
        shadow.plot_shadow(ax=ax)

    if title is not None:
        plt.suptitle(title)

    plt.tight_layout()
    if plot_dir is not None:
        plt.savefig(f"{plot_dir}{title}.pdf")
    else:
        plt.show()



def plot_density(xyz, gridsize=40, outname=None, **kwargs):
    """ Generate hexbin plots of the x-y and x-z particle densities.

    Args:
        gridsize (int): Number of bins in the hexbin plots.
        outname (str): Filename for saving the output plot. Will just run plt.show() if None.
    """
    figsize = kwargs.get("figsize", (8, 5))
    Rmax = kwargs.get("Rmax", 15)
    zmax = kwargs.get("zmax", 3)
    cmap = kwargs.get("cmap", "inferno")

    x, y, z = xyz
    lim = (-Rmax * 1.1, Rmax * 1.1)
    gridsize_z = int(gridsize * zmax / Rmax)

    fig, ax = plt.subplots(1, 2, figsize=figsize)

    hist1 = ax[0].hexbin(x, y, bins="log", cmap=cmap, gridsize=gridsize)
    hist2 = ax[1].hexbin(x, z, bins="log", cmap=cmap, gridsize=(gridsize, gridsize_z))
    
    for axis in ax.flatten():
        axis.set(xlim=lim, ylim=lim)

    ax[0].set_xlabel("x [kpc]")
    ax[0].set_ylabel("y [kpc]")
    ax[1].set_xlabel("x [kpc]")
    ax[1].set_ylabel("z [kpc]")

    plt.colorbar(mappable=hist1, ax=ax[0], location="top")
    plt.colorbar(mappable=hist2, ax=ax[1], location="top")

    plt.tight_layout()
    
    if outname is not None:
        plt.savefig(outname)
    else:
        plt.show()


def plot_density_3ax(xyz, gridsize=40, outname=None, **kwargs):
    """ Generate hexbin plots for all 3 principle axes of the particle densities.

    Args:
        gridsize (int): Number of bins in the hexbin plots.
        outname (str): Filename for saving the output plot. Will just run plt.show() if None.
    """
    figsize = kwargs.get("figsize", (12, 5))
    Rmax = kwargs.get("Rmax", 15)
    zmax = kwargs.get("zmax", 3)
    cmap = kwargs.get("cmap", "inferno")

    x, y, z = xyz
    lim = (-Rmax * 1.1, Rmax * 1.1)
    gridsize_z = int(gridsize * zmax / Rmax)

    fig, ax = plt.subplots(1, 3, figsize=figsize)

    hist1 = ax[0].hexbin(x, y, bins="log", cmap=cmap, gridsize=gridsize)
    hist2 = ax[1].hexbin(x, z, bins="log", cmap=cmap, gridsize=(gridsize, gridsize_z))
    hist3 = ax[2].hexbin(y, z, bins="log", cmap=cmap, gridsize=(gridsize, gridsize_z))

    for axis in ax.flatten():
        axis.set(xlim=lim, ylim=lim)

    ax[0].set_xlabel("x [kpc]")
    ax[0].set_ylabel("y [kpc]")
    ax[1].set_xlabel("x [kpc]")
    ax[1].set_ylabel("z [kpc]")
    ax[2].set_xlabel("y [kpc]")
    ax[2].set_ylabel("z [kpc]")

    plt.colorbar(mappable=hist1, ax=ax[0], location="top")
    plt.colorbar(mappable=hist2, ax=ax[1], location="top")
    plt.colorbar(mappable=hist3, ax=ax[2], location="top")

    plt.tight_layout()
    
    if outname is not None:
        plt.savefig(outname)
    else:
        plt.show()


from .. import general_plots as genplots

import galarp as grp

class TestDensityPlots:
    pot = grp.builtins.JZ2023_Satellite()
    mass_profile = grp.gen_mass_profile(pot)
    particles = grp.ExponentialGrid(n_particles=200)
    particles.generate(mass_profile=mass_profile)

    def test_2ax_plot(self, tmp_path):
        d = tmp_path / "sub"
        d.mkdir()
        genplots.plot_density(self.particles.get_xyz(), gridsize=10, outname=f'{d}_2ax_test.png')
    
    def test_3ax_plot(self, tmp_path):
        d = tmp_path / "sub"
        d.mkdir()
        genplots.plot_density_3ax(self.particles.get_xyz(), gridsize=10, outname=f'{d}_3ax_test.png')



from astropy import units as u
import numpy as np

from astropy.table import Table
from scipy.interpolate import interp1d


__all__ = [
    "Density",
    "ExponentialDensity",
    "InterpolatedDensity",
]

class Density:
    """Parent class to represent time-variable densities."""

    def __init__(self, rho):
        self.rho = rho
        assert type(rho) == u.Quantity, "Density must be a Quantity"

    def evaluate(self, t):
        return self.rho.to(u.g / u.cm**3)
    
    def evaluate_arr(self, ts):
        return np.array([self.evaluate(t).value for t in ts])


class ExponentialDensity(Density):
    """Class to represent an exponential density profile."""
    def __init__(self, rho, t0, width):
        super().__init__(rho)
        self.t0 = t0
        self.width = width

        assert type(t0) == u.Quantity, "t0 must be a Quantity"
        assert type(width) == u.Quantity, "width must be a Quantity"

    def evaluate(self, t):
        return self.rho * np.exp(-(((t - self.t0) / self.width) ** 2))


class InterpolatedDensity(Density):
    """Class to represent an interpolated density profile."""
    def __init__(self, interp=None, units=u.g/u.cm**3, **kwargs):
        super().__init__(rho=0 * u.g/u.cm**3, **kwargs)
        self.interp = interp

    def evaluate(self, t):
        """ Return the density at time t. """
        return self.interp(t) * u.g /u.cm **3

    @staticmethod
    def from_table(fn, t_key, rho_key,
                    t_units=u.s, rho_units=u.g/u.cm**3,
                    format="ascii", **kwargs):
        """
        Create an InterpolatedDensity object from a table.

        Parameters:
        fn (str): The filename of the table.
        t_key (str): The column name for the time values.
        rho_key (str): The column name for the density values.
        t_units (astropy.units.Unit, optional): The units for the time values. 
            Default is seconds (u.s).
        rho_units (astropy.units.Unit, optional): The units for the density values. 
            Default is grams per cubic centimeter (u.g/u.cm**3).
        format (str, optional): The format of the table. Default is "ascii".
        **kwargs: Additional keyword arguments to be passed to the InterpolatedDensity constructor.

        Returns:
        InterpolatedDensity: An InterpolatedDensity object created from the table.
        """
        t = Table.read(fn, format=format)

        ts = t[t_key] * t_units.to(u.Myr)

        rho = t[rho_key] * rho_units

        interp = interp1d(ts, rho, bounds_error=False, fill_value="extrapolate")

        return InterpolatedDensity(interp=interp, **kwargs)


import numpy as np
from astropy import units as u
from matplotlib import pyplot as plt

from .postprocessing.plotting import general_plots
import gala.dynamics as gd
from . import utils


from tqdm import tqdm


__all__ = [
    "ParticleGrid",
    "UniformGrid",
    "ExponentialGrid",
    "generate_exponential_positions",
]


class ParticleGrid:
    def __init__(self, name="PG"):
        self.container = []  # List of PhaseSpacePosition objects
        self.name = name

    def generate(self):
        raise NotImplementedError(
            "This is an abstract class. Please use a \
            subclass."
        )

    def plot_phase_space(self, ax=None, color="black", outname=None):
        if ax is None:
            fig, ax = plt.subplots(1, 2, facecolor="white", figsize=(8, 4))
        for particle in self.container:
            ax[0].scatter(particle.x, particle.y, color=color)
            ax[1].scatter(particle.v_x, particle.v_y, color=color)

        plt.tight_layout()
        if outname is not None:
            plt.savefig(outname, dpi=200)
        else:
            plt.show()
    
    def get_xyz(self):
        x, y, z = [], [], []
        for particle in self.container:
            x.append(particle.x.value)
            y.append(particle.y.value)
            z.append(particle.z.value)
        return x, y, z


    def plot_grid(self, **kwargs):
        """ Plot the grid of particles in the xy and xz planes. """
        particle_size = kwargs.get("particle_size", 1)

        fig, ax = plt.subplots(1, 2, facecolor="white", figsize=(8, 4))
        xs, ys, zs = self.get_xyz()

        ax[0].scatter(xs, ys, color="black", s=particle_size)
        ax[1].scatter(xs, zs, color="black", s=particle_size)
        
        ax[0].set_xlabel("x [kpc]")
        ax[0].set_ylabel("y [kpc]")
        ax[1].set_xlabel("x [kpc]")
        ax[1].set_ylabel("z [kpc]")
        
        ax[0].set_xlim(-self.Rmax.value, self.Rmax.value)
        ax[0].set_ylim(-self.Rmax.value, self.Rmax.value)
        ax[1].set_xlim(-self.Rmax.value, self.Rmax.value)
        ax[1].set_ylim(-self.Rmax.value, self.Rmax.value)

        plt.tight_layout()
        plt.show()
    

    def plot_density(self, gridsize=40, outname=None, **kwargs):
        """ Generate hexbin plots of the x-y and x-z particle densities.

        Args:
            gridsize (int): Number of bins in the hexbin plots.
            outname (str): Filename for saving the output plot. Will just run plt.show() if None.
        """
        general_plots.plot_density(self.get_xyz(), gridsize=gridsize, outname=outname,  **kwargs)


    def save(self, outname):
        utils.pickle_obj(self.container, outname)


class UniformGrid(ParticleGrid):
    """Class to initialize a grid of particles in the xy plane with circular
    velocities based on a given gravitational potential.

    Usage:
        ```
        mass_profile = gn.gen_mass_profile(potential)
        grid = gn.ParticleGrid()
        grid.generate(mass_profile)
        ```
    """

    def __init__(self, Rmax=10, n_particles=50, z_start=0.0, veldisp=10.0):
        """Generate a grid of particles in the xy plane with circular
            velocities based on a given gravitational potential.
            Can also include a Gaussian velocity dispersion (helpful for
            including a scale height in the disk).

        Args:
            mass_profile (interp1d): The mass profile of the potential in scipy
            interp1d format. Defaults to None.
            Rmax (_type_): The maximum radius of the grid particles.
            n_particles (int, optional): Number of particles between -Rmax and
                Rmax. Defaults to 50.
            z_start (float, optional): The z-level to initialize the grid in.
                Defaults to 0.
            veldisp (float, optional): _description_. Defaults to 10.
            velocities (bool, optional): If false, particles have no velocities.
                Defaults to True.
        """

        super().__init__(name="UG")

        self.Rmax = Rmax
        self.n_particles = n_particles
        self.z_start = z_start
        self.veldisp = veldisp

    def generate(self, mass_profile, velocities=True):
        particle_range = np.linspace(-self.Rmax, self.Rmax, self.n_particles) * u.kpc

        for x in particle_range:
            for y in particle_range:
                p_x0 = u.Quantity([x, y, self.z_start * u.kpc])
                R = np.sqrt(x**2 + y**2)
                if R > self.Rmax * u.kpc:
                    continue

                if velocities:
                    theta = np.arctan2(y, x)

                    # Initialize with circular velocity
                    p_v0 = utils.velocity(mass_profile, R, theta)
                    if self.veldisp is not None:
                        # Add random velocity dispersion
                        p_v0 += np.random.normal(scale=self.veldisp, size=3) * (
                            u.km / u.s
                        )
                else:
                    p_v0 = [0.0, 0.0, 0.0] * (u.km / u.s)

                w0 = gd.PhaseSpacePosition(pos=p_x0, vel=p_v0.to(u.kpc / u.Myr))
                self.container.append(w0)


class ExponentialGrid(ParticleGrid):
    def __init__(
        self,
        h_R=4 * u.kpc,
        h_z=0.5 * u.kpc,
        n_particles=500,
        veldisp=10.0 * u.km / u.s,
        Rmax=None,
        zmax=None,
    ):
        super().__init__(name="EG")

        self.h_R = h_R
        self.h_z = h_z

        self.Rmax = Rmax if Rmax is not None else self.h_R * 4
        self.zmax = zmax if zmax is not None else self.h_z * 4

        self.n_particles = n_particles
        self.veldisp = veldisp

    def generate(self, mass_profile, velocities=True, positions=None):
        if positions is None:  # Generate positions if nothing given
            xs, ys, zs = generate_exponential_positions(
                h_R=self.h_R,
                h_z=self.h_z,
                n_particles=self.n_particles,
                Rmax=self.Rmax,
                zmax=self.zmax,
                outname=None,
            )
        elif isinstance(positions, str):    # If string assume this is a filename
            xs, ys, zs = self.load_positions(positions)
        else:
            xs, ys, zs = positions

        # Create PhaseSpacePositions for all points, and initialize velocities
        for i in range(self.n_particles):
            x, y, z = xs[i], ys[i], zs[i]

            p_x0 = u.Quantity([x, y, z] * u.kpc)

            R = np.sqrt(x**2 + y**2 + z**2) * u.kpc

            if velocities:
                theta = np.arctan2(y, x)
                vx, vy = self.velocity(
                    mass_profile, R, theta
                )  # Initialize with circular velocity

                inc = np.arctan2(z, np.sqrt(x**2 + y**2))
                vtot = np.sqrt(vx**2 + vy**2)
                vz = (
                    vtot * np.sin(inc)
                )  # TODO improve this (better than nothing but not ideal. Leads to "sloshing")

                p_v0 = u.Quantity([vx, vy, vz])

                if self.veldisp is not None:
                    p_v0 += np.random.normal(scale=self.veldisp.value, size=3) * (
                        u.km / u.s
                    )  # Add random velocity dispersion
            else:
                p_v0 = [0.0, 0.0, 0.0] * (u.km / u.s)

            w0 = gd.PhaseSpacePosition(pos=p_x0, vel=p_v0.to(u.kpc / u.Myr))
            self.container.append(w0)

    def load_positions(self, filename):
        """Create the grid of particles from a file of positions (to save on time not needed to rerun the MC sim)
        Args:
            filename (numpy savetxt file): Text file of positions
        """
        positions = np.load(filename)
        return positions

    def velocity(self, mass_profile, R, theta):
        """ Get the velocity of a particle in the xy plane given an angle and the mass profile
        
        Args:
            mass_profile (interp1d): The mass profile of the potential in scipy interp1d format.
            R (float): The radius of the particle.
            theta (float): The angle of the particle.
        """
        vx = -utils.v_circ(mass_profile(R) * u.M_sun, R) * np.sin(theta)
        vy = utils.v_circ(mass_profile(R) * u.M_sun, R) * np.cos(theta)
        return vx, vy


def generate_exponential_positions(
    n_particles,
    h_R=4 * u.kpc,
    h_z=0.5 * u.kpc,
    Rmax=15 * u.kpc,
    zmax=2 * u.kpc,
    outname=None,
):
    def remap(value, a, b):
        return (value * (b - a)) + a

    n_0 = (
        1 / (4 * np.pi * h_R**2 * h_z).value
    )  # This normalizes the probability distribution

    h_R, h_z = h_R.to(u.kpc).value, h_z.to(u.kpc).value
    rmax = Rmax.to(u.kpc).value
    zmax = zmax.to(u.kpc).value

    Rs = []
    zs = []

    for n in tqdm(range(n_particles)):
        good = False
        while not good:  # TODO maybe add a max_tries here with a warning message
            rand_R, rand_z = np.random.random(2)
            R, z = remap(rand_R, 0, rmax), remap(rand_z, -zmax, zmax)

            v = n_0 * np.exp(-R / h_R) * np.exp(-np.abs(z) / h_z)

            if np.random.random() <= v:
                Rs.append(R)
                zs.append(z)

                good = True

    thetas = np.random.random(len(Rs)) * (2 * np.pi)
    xs, ys = Rs * np.cos(thetas), Rs * np.sin(thetas)

    if outname is not None:
        np.save(outname, np.array([xs, ys, zs]))

    return xs, ys, zs



from .helpers import ParticleGridTestBase
from .. import particle_grids as pgrids

from ..builtins.satellites import JZ2023_Satellite
from ..utils import gen_mass_profile


class TestUniformGrid(ParticleGridTestBase):
    name = "UniformGrid"
    grid = pgrids.UniformGrid(Rmax=10, n_particles=50, z_start=0.0, veldisp=10.0)
    

    example_potential = JZ2023_Satellite()
    mass_profile = gen_mass_profile(example_potential)

    grid.generate(mass_profile, velocities=True)


class TestExpontentialGrid(ParticleGridTestBase):
    name = "ExponentialGrid"
    grid = pgrids.ExponentialGrid(n_particles=200)
    

    example_potential = JZ2023_Satellite()
    mass_profile = gen_mass_profile(example_potential)

    grid.generate(mass_profile)


# Testing utilities


# Third Party
import os
import numpy as np


# This module
import galarp as grp

from gala.units import galactic
import astropy.units as u


__all__ = ["test_uniform_shadow"]


def test_uniform_shadow(plot_dir="plots/shadow_tests/"):
    """Test that the angled shadows being created are correct"""

    os.makedirs(plot_dir, exist_ok=True)

    for angle in range(10, 90, 10):
        wind = grp.RPWind(units=galactic)
        wind.init_from_inc(np.deg2rad(angle), 300 * u.km / u.s)

        shadow = grp.UniformShadow()
        shadow.init_from_wind(wind)
        shadow.plot_shadow(wind=wind, outname=f"{plot_dir}shadow_{angle}.pdf")




import numpy as np

import matplotlib as mpl

__all__ = ["pyplot_style", "default_pyplot_style", "lavender_cmap"]
    

def pyplot_style():
    """ If you just call pyplot style, return the default. """
    default_pyplot_style()


def default_pyplot_style():
    # xticks
    mpl.rcParams["xtick.major.size"] = 3
    mpl.rcParams["xtick.major.width"] = 2
    mpl.rcParams["xtick.minor.size"] = 2
    mpl.rcParams["xtick.minor.width"] = 1

    # yticks
    mpl.rcParams["ytick.major.size"] = 3
    mpl.rcParams["ytick.major.width"] = 2
    mpl.rcParams["ytick.minor.size"] = 2
    mpl.rcParams["ytick.minor.width"] = 1

    mpl.rcParams["axes.linewidth"] = 1.5

    mpl.rc("xtick", labelsize=12)
    mpl.rc("ytick", labelsize=12)

    font = {
        "family": "serif",
        #'weight': 'bold',
        "size": 12,
    }

    mpl.rc("font", **font)

    mpl.rc("lines", linewidth=1, linestyle="solid", color="black")


def lavender_cmap(step_1=175):
    x, y = np.mgrid[:251, :251]

    lavender = [86, 82, 100]
    indigo = [29, 0, 51]
    yellow = [100, 100, 0]

    step_2 = 256 - step_1

    r, g, b = [], [], []

    r = np.array(np.append(np.linspace(lavender[0], indigo[0], step_1), 
                     np.linspace(indigo[0], yellow[0], step_2))) / 100
    g = np.array(np.append(np.linspace(lavender[1], indigo[1], step_1), 
                     np.linspace(indigo[1], yellow[1], step_2))) / 100
    b = np.array(np.append(np.linspace(lavender[2], indigo[2], step_1), 
                     np.linspace(indigo[2], yellow[2], step_2))) / 100

    a = np.array([1 for i in range(len(r))])

    full_arr = np.asarray([[r[i], g[i], b[i], a[i]] for i in range(0, len(r))])

    new_cmap = mpl.colors.ListedColormap(full_arr)
    return new_cmap




from astropy.table import Table
import astropy.units as u
from astropy.units import Quantity
import numpy as np

from gala.units import galactic

from scipy.interpolate import interp1d



__all__ = [
    "RPWind",
    "LorentzianWind",
    "StepFunctionWind",
    "InterpolatedWind",
]


class RPWind:
    """Class to represent a ram pressure wind

    Usage:
        ```
        wind = gn.RPWind([50, 0, 50] * (u.km / u.s), units=gn.galactic)  # Initializes a wind at 45 degrees in the x-z plane.

        wind = gn.RPWind(units=galactic)                                 # Achieves the same thing.
        wind.init_from_inc(np.deg2rad(45), 100 * u.km  / u.s)
        ```
    """

    def __init__(self, vector=None, units=None, **kwargs):
        """
        Args:
            vector (astropy Quantity, optional): Wind vector. Can also be initialized using self.init_from_inc(). Defaults to None.
            units (_type_, optional): _description_. Defaults to None.
        """
        if vector is None:
            self.vector = Quantity([0, 0, 0]) * (u.km / u.s)
        # self.vector = Quantity(vector)

        self.units = units
        if self.units is not None:
            if type(self.vector) == Quantity:
                self.vector = self.vector.to(self.units["length"] / self.units["time"])
            else:
                self.vector *= self.units["length"] / self.units["time"]

    def evaluate(self, t):
        """Return the wind vector at time t. For the default wind vector, this is JUST the vector converted to kpc/Myr"""
        return self.vector.to(u.kpc / u.Myr).value
    
    def evaluate_arr(self, ts):
        return np.array([self.evaluate(t) for t in ts])

    def initialize_vector(self):
        if self.units is not None:
            self.vector = self.vector.to(self.units["length"] / self.units["time"])

    def init_from_inc(self, inclination, strength):
        """Initialize the wind vector from an inclination and strength"""

        x = strength * np.cos(inclination)
        z = strength * np.sin(inclination)
        self.vector = Quantity([x, 0 * x.unit, z]).to(
            self.units["length"] / self.units["time"]
        )

    def wind_strength(self):
        """Return the length (strength) of the wind vector"""
        return np.sqrt(sum(self.vector**2))

    def normalized(self):
        """Return the normalized wind vector"""
        return self.vector / self.wind_strength()

    def vector_to_units(self, units):
        return self.vector.to(units)

    def vector_as_value(self):
        if type(self.vector) == Quantity:
            return self.vector.value
        else:
            return self.vector

    def inclination(self):
        x, y, z = self.vector_as_value()
        return np.arctan2(z, np.sqrt(x**2 + y**2))

    def __repr__(self):
        return f"<RP Wind Vector={self.vector}  Inclination={np.round(np.rad2deg(self.inclination()), 2):.2f}  >"


class LorentzianWind(RPWind):
    """Ram pressure wind that is damped by a Lorentzian profile.
        Reaches max (which is just the unadjusted RP wind value) at t0.

    Args:
        gn (_type_): _description_
    """

    def __init__(self, t0=0 * u.Myr, width=200 * u.Myr, **kwargs):
        super().__init__(**kwargs)
        self.t0 = t0.to(u.Myr).value  # Units are in Myr
        self.width = width.to(u.Myr).value  # Units are in Myr

    def evaluate(self, t):
        """Return the wind vector damped by a Lorentzian profile"""
        return super().evaluate(t) * 1 / ((2 * (t - self.t0) / self.width) ** 2 + 1)


class StepFunctionWind(RPWind):
    """A wind represented by a step function (no wind immediately into full wind at time t0)
    Good for testing that time variability is working.

    """

    def __init__(self, t0=0 * u.Myr, **kwargs):
        super().__init__(**kwargs)
        self.t0 = t0.to(u.Myr).value

    def evaluate(self, t):
        factor = t > self.t0
        return super().evaluate(t) * factor
    

class InterpolatedWind(RPWind):
    """Wind that is interpolated from a table of values. """

    def __init__(self, interp=None, units=galactic, **kwargs):
        """
            Args:
                interp (callable): Interpolation function that takes a time value as input and returns the wind velocity at that time.
                units (dict, optional): Dictionary specifying the units of the wind vector. Defaults to galactic units.
                inc (float, optional): Inclination angle of the wind vector in radians. Defaults to 0.
            """
        super().__init__(**kwargs)
        self.interp = interp
        self.inc = kwargs.get("inc", 0)

        self.unit_vector = [np.cos(self.inc), 0, np.sin(self.inc)]
    
    def evaluate(self, t):
        """ Evaluate the wind vector at a given time by multiplying the unit vector by the interpolated
            wind strength at that time.
        """
        return self.unit_vector * self.interp(t)

    @staticmethod
    def from_table(fn, t_key, vel_keys, format="ascii",
                   t_units = u.s, v_units = u.cm / u.s, 
                   verbose=False, **kwargs):
        """
        Create an InterpolatedWind object from a table.

        Parameters:
        - fn (str): The filename of the table.
        - t_key (str): The key for the time column in the table.
        - vel_keys (str or list): The key(s) for the velocity column(s) in the table.
        - format (str, optional): The format of the table. Default is "ascii".
        - t_units (astropy.units.Unit, optional): The units for the time column. Default is u.s.
        - v_units (astropy.units.Unit, optional): The units for the velocity column(s). Default is u.cm / u.s.
        - verbose (bool, optional): Whether to print verbose output. Default is False.
        - **kwargs: Additional keyword arguments to be passed to the InterpolatedWind constructor.

        Returns:
        - InterpolatedWind: An InterpolatedWind object created from the table.
        Usage:
                >>> wind = InterpolatedWind.from_table("wind_data.txt", "time", 
                    ["velocity_x", "velocity_y", "velocity_z"], units=galactic, inc=np.deg2rad(45))
                >>> print(wind.evaluate(10 * u.Myr))
        """
        t = Table.read(fn, format=format)

        if verbose:
            print(f'Loaded table with {len(t)} rows and keys: {t.keys()}')

        ts = t[t_key] * t_units.to(u.Myr)

        if not isinstance(vel_keys, list): vel_keys = [vel_keys]

        vels = np.array([t[key] for key in vel_keys])
        v_tot = np.sqrt(np.sum(vels**2, axis=0)) * v_units.to(u.kpc/u.Myr)

        interp = interp1d(ts, v_tot, bounds_error=False, fill_value="extrapolate")

        return InterpolatedWind(interp=interp, **kwargs)



"""Classes for implementing orbits under ram pressure acceleration."""

# Third Party
import os
import pickle
import astropy.units as u
import numpy as np

from .postprocessing import utils

# This module
from . import shadows, winds, densities

import gala.dynamics as gd
from gala.units import galactic


# import integrate as gi            # If debugging
import gala.integrate as gi


__all__ = ["F_RPS", "RPSim", "OrbitContainer"]


def F_RPS(
    t, w, potential, shadow, wind, rho, r_cloud, m_cloud, wind_on=True, debug=False
):
    # position units are in kpc
    # velocity units here are in kpc/Myr
    x, y, z, vx, vy, vz = w
    q = np.stack((x, y, z), axis=1)
    p = np.stack((vx, vy, vz), axis=1)

    # compute acceleration from potential:
    _t = np.array([0.0])
    acc = -potential._gradient(q, _t).T

    # Compute acceleration from ram pressure
    # a_ram = pi * rho * v_perp^2 * r_cloud^2 / m_cloud
    # TODO remove units altogether
    if wind_on:
        v_perp = p - wind.evaluate(t)
        a_ram = (np.pi * rho.evaluate(t) * r_cloud**2 / m_cloud).to(1 / u.kpc).value * (
            v_perp**2
        )

        a_ram = a_ram.T

        # If wind is on and shadow exists, apply shadow to appropriate particles
        if shadow is not None:
            shadow = shadow.evaluate(q, t).T
            a_ram *= shadow

        acc += a_ram

    return np.vstack((p.T, acc))


class RPSim:
    def __init__(
        self,
        wind,
        potential,
        rho_icm=None,
        shadow=None,
        potential_name="",
        method=F_RPS,
    ):
        self.method = method

        self.wind = wind
        self.potential = potential

        self.shadow = shadow

        self.potential_name = potential_name

        self.rho_icm = rho_icm

        self.inclination = wind.inclination()
        if (
            np.rad2deg(self.inclination) < 10
            and type(self.shadow) != shadows.EdgeOnShadow
        ):
            print(
                "Warning: Inclination is less than 10 degrees. You should be using an EdgeOnShadow."
            )

        self.sim_results = []

    def run(
        self,
        particles,
        rho_icm=2e-27 * (u.g / u.cm**3),
        t0=100 * u.Myr,
        r_cloud=50 * u.pc,
        m_cloud=1e5 * u.Msun,
        wind_on=True,
        integration_time=500 * u.Myr,
        dt=5 * u.Myr,
        printout=True,
        wind=None,
        outdir=None,
        debug=False,
    ):

        if type(rho_icm) is u.Quantity:
            self.rho_icm = densities.Density(rho_icm)
        else:
            self.rho_icm = rho_icm

        # Allow for user to switch out wind in the run method
        if wind is not None:
            self.wind = wind

        if printout:
            printout_width = 80
            print("".center(printout_width, "-"))
            print(f" Running GALA sim with  {self.wind.wind_strength():.2e}  wind at  {self.wind.inclination():.2f}  degrees ".center(
                    printout_width, "-"))
            print(f" Running for {integration_time}  at a timestep of  {dt}  ({integration_time / dt:.1f} steps) ".center(
                    printout_width, "-"))
            print("".center(printout_width, "-"))
            print(f" Container contains {len(particles.container)} particles ".center(
                    printout_width, "-"))
            print(f" Each cloud has a mass of  {m_cloud.value:.1e} solar masses  and a radius of  {r_cloud:.2f}   ".center(
                    printout_width, "-"))
            print(f" This corresponds to a density of {(m_cloud / r_cloud **3).to(u.g/u.cm**3):.2e}   ".center(
                    printout_width, "-"))
            print("".center(printout_width, "-"))

        # potential, shadow, wind_vec, rho, r_cloud, m_cloud, damping=1, wind_on=True)

        integrator = gi.RK5Integrator(
            self.method,
            func_args=(
                self.potential,
                self.shadow,
                self.wind,
                self.rho_icm,
                r_cloud,
                m_cloud,
                wind_on,
            ),
            func_units=galactic,
            progress=not debug,
        )

        orbits = integrator.run(
            gd.combine(particles.container), dt=dt, t1=0, t2=integration_time
        )
        self.sim_results.append(orbits)

        metadata = {
            "WIND": self.wind,
            "POTENTIAL": self.potential,
            "POTENTIAL_NAME": self.potential_name,
            "SHADOW": self.shadow,
            "RHO_ICM": self.rho_icm,
            "PARTICLE_GRID": particles,
            "R_CLOUD": r_cloud,
            "M_CLOUD": m_cloud,
            "WIND_ON": wind_on,
            "INTEGRATION_TIME": integration_time,
            "DT": dt,
        }

        out = OrbitContainer(orbits, units=galactic, metadata=metadata)

        if outdir is not None:
            os.makedirs(outdir, exist_ok=True)
            save_orbits(out, outdir=outdir)

        return out

    def __repr__(self) -> str:
        return f"Ram Pressure Sim \n Wind: \t\t{self.wind.__repr__()}\n Shadow: \t{self.shadow.__repr__()}\n Potential: \t{self.potential}"


def autogen_fn(md, suffix="orbits"):
    fn = ""

    if md["POTENTIAL_NAME"] != "":
        fn += f'{md["POTENTIAL_NAME"]}_'
    if md["WIND"] is not None:
        wind = md["WIND"]
        fn += f"WIND_{wind.wind_strength().to(u.km/u.s).value:.0f}_{np.rad2deg(wind.inclination()):.0f}_"
    if md["SHADOW"] is not None:
        shadow = md["SHADOW"]
        fn += f"_{shadow.shadow_name}_{shadow.damping}DAMP_"
    fn += f'{md["M_CLOUD"].value:.0e}Msun_{md["R_CLOUD"].value:.0f}pc'

    fn += f".{suffix}"

    return fn


def save_orbits(orbits, name="auto", outdir=""):
    metadata = orbits.metadata
    if name == "auto":
        name = autogen_fn(metadata)

    with open(outdir + name, "wb") as f:
        pickle.dump(orbits, f)


class OrbitContainer:
    """ Container for storing orbits and metadata from a completed GalaRP run.
    """
    def __init__(self, data, units=None, metadata={}):
        self.data = data
        self.units = units

        self.metadata = metadata

    def plot(self, plot_3d=False, plot_orbits=False):
        if plot_3d:
            utils.k3d_plot([self])

        if plot_orbits:
            utils.plot_orbits(
                self.data, wind=self.metadata["WIND"], shadow=self.metadata["SHADOW"]
            )





class ShadowTestBase:
    name = None
    shadow = None
    show_plots = False

    xyz = None
    t = None

    def test_evaluation(self):
        eval = self.shadow.evaluate(self.xyz, self.t)
        print(len(self.xyz), eval.shape)
        assert eval is not None
        assert eval.shape[0] == len(self.xyz)


class RPWindTestBase:
    name = None
    wind = None
    show_plots = False

    t = None

    def test_init(self):
        assert self.wind is not None


class ParticleGridTestBase:
    name = "ParticleGrids"
    grid = None
    show_plots = False

    def test_init(self):
        assert self.grid is not None
        assert isinstance(self.grid.container, list) 
    
    



import numpy as np
from matplotlib import pyplot as plt

import matplotlib.patches as patches



#############################################
######## Plotting routines ##################
#############################################


def ellipse_coords(x, y, a, b, theta, num_points=100, b_is_ellipticity=False):
    # Generate angles for sampling
    angles = np.linspace(0, 2 * np.pi, num_points)

    # If b is supplied as an ellipticity (and not as the semiminor axis directly, calculate semiminor axis)
    if b_is_ellipticity:
        b = a * (1 - b)

    # Parametric equation for the ellipse
    x_coords = (
        x + a * np.cos(angles) * np.cos(theta) - b * np.sin(angles) * np.sin(theta)
    )
    y_coords = (
        y + a * np.cos(angles) * np.sin(theta) + b * np.sin(angles) * np.cos(theta)
    )
    z_coords = np.zeros(num_points)
    # Return the sampled coordinates as a NumPy array
    return x_coords, y_coords, z_coords


def normalize_vector(vec):
    return vec / np.sqrt(np.sum(vec**2))


def plot_axis_symbol(ax, val, x0, y0, color, length):
    if abs(val) > 1e-1:
        if val > 0:
            ax.text(
                x0,
                y0,
                r"$\odot$",
                ha="center",
                va="center",
                size=length * 3,
                color=color,
            )
        else:
            ax.text(
                x0,
                y0,
                r"$\otimes$",
                ha="center",
                va="center",
                size=length * 3,
                color=color,
            )


def _limits_and_span(ax):
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    xspan, yspan = xmax - xmin, ymax - ymin
    return xmin, xmax, xspan, ymin, ymax, yspan


def plot_wind_vector(vector, ax, length=5, loc=(0, 0, 0), color="black"):
    vector = normalize_vector(vector) * length
    length = 5

    x0, y0, z0 = loc

    dx, dy, dz = vector * length

    xmin, xmax, xspan, ymin, ymax, yspan = _limits_and_span(ax[0])

    if abs(dx) > 1e-1 or abs(dy) > 1e-1:
        ax[0].arrow(x0, y0, dx, dy, head_width=1, color=color, zorder=20)
    else:
        plot_axis_symbol(ax[0], dy, x0, y0, color=color, length=length)

    xmin, xmax, xspan, ymin, ymax, yspan = _limits_and_span(ax[1])

    if abs(dx) > 1e-1 or abs(dz) > 1e-1:
        ax[1].arrow(x0, z0, dx, dz, head_width=1, color=color, zorder=20)
    else:
        plot_axis_symbol(ax[1], dy, x0, z0, color=color, length=length)

    xmin, xmax, xspan, ymin, ymax, yspan = _limits_and_span(ax[2])
    if abs(dy) > 1e-1 or abs(dz) > 1e-1:
        ax[2].arrow(y0, z0, dy, dz, head_width=1, color=color, zorder=20)
    else:
        plot_axis_symbol(ax[2], dx, y0, z0, color=color, length=length)


def plot_rotation_vector(ax, color="red", clockwise=False):
    xmin, xmax, xspan, ymin, ymax, yspan = _limits_and_span(ax)
    x0, y0 = xmax - xspan / 4, ymin + yspan / 5

    if clockwise:
        dx, dy = -xspan / 5, -yspan / 5
    else:
        dx, dy = xspan / 5, xspan / 5

    style = "Simple, tail_width=0.8, head_width=8, head_length=8"
    kw = dict(arrowstyle=style, color=color)
    a3 = patches.FancyArrowPatch(
        (x0, y0), (x0 + dx, y0 + dy), zorder=20, connectionstyle="arc3,rad=.25", **kw
    )
    ax.add_patch(a3)


def plot_disk(ax, R, lw=1, color="black"):
    """Plot a disk of a given radius in a 3-panel plot"""
    ell_xs, ell_ys, ell_zs = ellipse_coords(0, 0, R, R, 0)
    ax[0].plot(ell_xs, ell_ys, lw=lw, color=color)
    ax[1].plot(ell_xs, ell_zs, lw=lw, color=color)
    ax[2].plot(ell_ys, ell_zs, lw=lw, color=color)


def diagnostic_3ax(rpsim, length=3, wind_vec_loc=(0, 0, 0), outname=None):
    """Diagnostic plot for 3D wind vector and its projections onto the xy, xz, and yz planes.

    Args:
        rpsim (_type_): _description_
        length (int, optional): _description_. Defaults to 3.
        wind_vec_loc (tuple, optional): _description_. Defaults to (0,0,0).
    """

    fig, ax = plt.subplots(1, 3, facecolor="white", figsize=(12, 4))

    ell_xs, ell_ys, ell_zs = ellipse_coords(0, 0, 10, 10, 0, 100)

    ax[0].plot(ell_xs, ell_ys, color="black", lw=1)
    ax[1].plot(ell_xs, ell_zs, color="black", lw=1)
    ax[2].plot(ell_ys, ell_zs, color="black", lw=1)

    vector = rpsim.wind.normalized().value

    plot_wind_vector(vector, ax, length=length, loc=wind_vec_loc)

    xlabels, ylabels = ["x", "x", "y"], ["y", "z", "z"]

    for i, axis in enumerate(ax.flatten()):
        axis.set_xlim(-12, 12)
        axis.set_ylim(-12, 12)
        axis.set_xlabel(xlabels[i])
        axis.set_ylabel(ylabels[i])

    plt.tight_layout()
    if outname is not None:
        plt.savefig(outname, dpi=200)
    else:
        plt.show()



import numpy as np
from astropy import units as u
from matplotlib import pyplot as plt

from . import utils

__all__ = ["ShadowBase", "UniformShadow", "ExponentialShadow", "EdgeOnShadow"]


def _shadow_tangent(z, phi):
    return z / np.tan(phi)


class ShadowBase:
    """
    Base class for creating shadow objects.

    Parameters:
    - damping (float): The damping factor for the shadow.
    - R_disk (Quantity or float): The radius of the shadow disk.
    - shadow_name (str): The name of the shadow.

    Methods:
    - evaluate(xyz): Evaluates the shadow at the given coordinates.
    - init_from_wind(wind): Initializes the shadow from a wind object.
    - plot_shadow(ax, wind, color, outname, x0, y0, z0): Plots the shadow on the given axes.
    - __repr__(): Returns a string representation of the shadow object.
    """

    def __init__(self, damping=0.5, R_disk=10 * u.kpc, shadow_name="ShadowBase"):
        self.damping = damping
        self.R_disk = R_disk
        self.shadow_name = shadow_name

    def evaluate(self, xyz, t):
        """
        Evaluates the shadow at the given coordinates.

        Parameters:
        - xyz (array-like): The coordinates at which to evaluate the shadow.

        Returns:
        - result (array-like): The evaluated shadow values.
        """
        raise NotImplementedError

    def init_from_wind(self, wind):
        """
        Initializes an angled shadow from a wind object.

        Parameters:
        - wind (Wind): The wind object from which to initialize the shadow.
        """
        self.phi = wind.inclination()

    def plot_shadow(
        self, ax=None, wind=None, color="black", outname=None, x0=0, y0=0, z0=0
    ):
        """
        Plots the shadow on the given axes.

        Parameters:
        - ax (array-like, optional): The axes on which to plot the shadow. If not provided, a new figure will be created.
        - wind (Wind, optional): The wind object to plot along with the shadow.
        - color (str, optional): The color of the shadow.
        - outname (str, optional): The name of the output file to save the plot.
        - x0, y0, z0 (float, optional): The coordinates of the origin of the shadow.

        Returns:
        - None
        """
        R_disk = self.R_disk.value if type(self.R_disk) is u.Quantity else self.R_disk
        R_plot = R_disk * 1.5

        # Create figure if ax is not provided
        if ax is None:
            fig, ax = plt.subplots(1, 3, facecolor="white", figsize=(12, 4))

        for axis in ax.flatten():
            axis.set_xlim(-R_plot, R_plot)
            axis.set_ylim(-R_plot, R_plot)

        x = np.linspace(-R_disk, R_disk, 100)
        y = np.linspace(-R_disk, R_disk, 100)
        X, Y = np.meshgrid(x, y)
        XY = self.evaluate(np.array([X, Y, np.zeros(X.shape) + z0]).T)
        XZ = self.evaluate(np.array([X, np.zeros(X.shape) + y0, Y]).T)
        YZ = self.evaluate(np.array([np.zeros(X.shape) + x0, X, Y]).T)

        im1 = ax[0].imshow(
            1 - XY,
            origin="lower",
            cmap="Greys",
            alpha=0.5,
            extent=(-R_disk, R_disk, -R_disk, R_disk),
        )
        im2 = ax[1].imshow(
            1 - XZ,
            origin="lower",
            cmap="Greys",
            alpha=0.5,
            extent=(-R_disk, R_disk, -R_disk, R_disk),
        )
        im3 = ax[2].imshow(
            1 - YZ,
            origin="lower",
            cmap="Greys",
            alpha=0.5,
            extent=(-R_disk, R_disk, -R_disk, R_disk),
        )

        plt.colorbar(mappable=im1, ax=ax[0], location="top")
        plt.colorbar(mappable=im2, ax=ax[1], location="top")
        plt.colorbar(mappable=im3, ax=ax[2], location="top")

        utils.plot_disk(ax, R_disk, lw=1, color="black")

        if wind is not None:
            utils.plot_wind_vector(
                wind.normalized().value, ax, length=0.5, loc=(-R_disk, -R_disk, -R_disk)
            )

        if ax is None:
            plt.tight_layout()
            if outname is not None:
                plt.savefig(outname, dpi=200)
            else:
                plt.show()

    def __repr__(self):
        """
        Returns a string representation of the shadow object.

        Returns:
        - repr (str): The string representation of the shadow object.
        """
        return f"<{self.shadow_name} RP Shadow:  Phi={np.rad2deg(self.phi):.2f}  Damping={self.damping}  R_Disk={self.R_disk}  zmin={self.zmin} >"


class UniformShadow(ShadowBase):
    """A class representing a uniform angled shadow.

    This class inherits from the ShadowBase class.

    Args:
        damping (float, optional): The damping factor. Defaults to 0.5.
        R_disk (Quantity, optional): The radius of the disk. Defaults to 10 kpc.
        zmin (Quantity, optional): The minimum value of z above the disk. Defaults to 0 kpc.
        phi (float, optional): The angle in radians. Defaults to 20 degrees.
    """

    def __init__(
        self, damping=0.5, R_disk=10 * u.kpc, zmin=0 * u.kpc, phi=np.deg2rad(20)
    ):
        super().__init__(damping=damping, R_disk=R_disk, shadow_name="Uniform")
        self.zmin, self.phi = zmin, phi

    def evaluate(self, q, t):
        x, y, z = q.T

        cent = _shadow_tangent(z, self.phi)
        dist = np.sqrt((x - cent) ** 2 + y**2)

        out = np.ones(dist.shape)
        in_disk = np.logical_and((z > self.zmin.value), (dist < self.R_disk.value))
        out[in_disk] = self.damping
        return out

    def plot_shadow(
        self, ax=None, wind=None, color="black", outname=None, x0=0, y0=0, z0=None
    ):
        z0 = self.zmin + 0.5 * u.kpc if z0 is None else z0
        super().plot_shadow(
            ax=ax, wind=wind, color=color, outname=outname, x0=z0, y0=y0, z0=z0
        )


class ExponentialShadow(ShadowBase):
    """A class representing an exponential drop-off angled shadow.

    This class inherits from the ShadowBase class.

    Args:
        damping (float, optional): The damping factor. Defaults to 0.5.
        R_disk (Quantity, optional): The radius of the disk. Defaults to 10 kpc.
        zmin (Quantity, optional): The minimum value of z above the disk. Defaults to 0 kpc.
        phi (float, optional): The angle in radians. Defaults to 20 degrees.
    """

    def __init__(
        self, damping=0.5, R_disk=10 * u.kpc, zmin=0 * u.kpc, phi=np.deg2rad(20)
    ):
        super().__init__(damping=damping, R_disk=R_disk, shadow_name="Exponential")
        self.zmin, self.phi = zmin, phi

    def evaluate(self, q, t):
        x, y, z = q.T
        cent = _shadow_tangent(z, self.phi)
        dist = np.sqrt((x - cent) ** 2 + y**2)
        out = np.exp(-dist / self.R_disk.value)
        in_disk = np.logical_and((z > self.zmin.value), (dist < self.R_disk.value))
        out[in_disk] *= self.damping
        return out

    def plot_shadow(
        self, ax=None, wind=None, color="black", outname=None, x0=0, y0=0, z0=None
    ):
        z0 = self.zmin + 0.5 * u.kpc if z0 is None else z0
        super().plot_shadow(
            ax=ax, wind=wind, color=color, outname=outname, x0=z0, y0=y0, z0=z0
        )


class EdgeOnShadow(ShadowBase):
    def __init__(self, damping=0.5, R_disk=10 * u.kpc, Z_disk=2 * u.kpc, x0=0 * u.kpc):
        super().__init__(damping=damping, R_disk=R_disk, shadow_name="EdgeOn")

        # Always assume kpc for xyz inputs (galactic coordinate system)
        self.x0 = x0.to(u.kpc).value
        self.R_disk = R_disk.to(u.kpc).value
        self.Z_disk = Z_disk.to(u.kpc).value

    def evaluate(self, q, t):
        x, y, z = q.T

        in_ellipsoid = (y / self.R_disk) ** 2 + (z / self.Z_disk) ** 2 < 1

        out = np.ones(x.shape)

        out[np.logical_and((x >= 0), in_ellipsoid)] = self.damping

        return out

    def plot_shadow(
        self, ax=None, wind=None, color="black", outname=None, x0=0, y0=0, z0=0
    ):
        super().plot_shadow(
            ax=ax, wind=wind, color=color, outname=outname, x0=z0, y0=y0, z0=z0
        )



import numpy as np

from .helpers import ShadowTestBase
from .. import shadows 

class TestUniformShadow(ShadowTestBase):
    shadow = shadows.UniformShadow()
    xyz = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    t = 0

class TestExponentialShadow(ShadowTestBase):
    shadow = shadows.ExponentialShadow()
    xyz = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    t = 0

class TestEdgeOnShadow(ShadowTestBase):
    shadow = shadows.EdgeOnShadow()
    xyz = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    t = 0
    



#!/usr/bin/env python
# Licensed under an MIT license - see LICENSE

from galarp import __version__


from setuptools import setup, find_packages


from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.rst").read_text()


setup(
    name='galarp',
    version=__version__,
    author='Harrison Souchereau',
    author_email='harrison.souchereau@yale.edu',
    description='A ram pressure add-on for Gala numerical integration of gravitational potentials',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/HSouch/galarp',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    keywords='galaxies, gravity, gravitational potentials',
    install_requires=[
        "astropy>=6.0.0",
        "gala>=1.8.1",
        "k3d>=2.16.1",
        "matplotlib>=3.8.3",
        "numpy>=1.26.4",
        "scipy>=1.13.0",
        "tqdm>=4.66.1",
    ],
)


import numpy as np

import galarp as grp

class TestRotate:
    """ Check that the rotation matrix function actually works as intended, including that it doesn't
        introduce any pesky values when rotating about a principle axis.
    """
    def test_rotate(self):
        # Test rotation of a vector
        xs, ys, zs = np.random.random(100), np.random.random(100), np.random.random(100)
        vec = np.array([xs, ys, zs])
        rotated = grp.rotate(vec, alpha=np.pi/4, beta=np.pi/4, gamma=np.pi/4)
        assert vec.shape == rotated.shape

    def test_rotate_about_axis(self):
        # Test rotation about the z-axis, which should not introduce values in the z-axis (which are all 
        # initialized to 0.)
        
        rand, zero_vec = np.random.random(100), np.zeros(100)

        rotated_x0 = grp.rotate(np.array([zero_vec, rand, rand]), alpha=np.pi/4)
        assert np.max(np.abs(rotated_x0[0])) < 1e-12

        rotated_y0 = grp.rotate(np.array([rand, zero_vec, rand]), beta=np.pi/4)
        assert np.max(np.abs(rotated_y0[1])) < 1e-12

        rotated_z0 = grp.rotate(np.array([rand, rand, zero_vec]), gamma=np.pi/4)
        assert np.max(np.abs(rotated_z0[2])) < 1e-12




import numpy as np
from astropy import units as u


from .helpers import RPWindTestBase
from ..winds import RPWind, LorentzianWind, StepFunctionWind, InterpolatedWind


class TestRPWind(RPWindTestBase):
    wind = RPWind()
    t = 0


class TestLorentzianWind(RPWindTestBase):
    wind = LorentzianWind(t0=500 * u.Myr)
    t = 0


class TestStepFunctionWind(RPWindTestBase):
    wind = StepFunctionWind(t0=500 * u.Myr)
    

class TestInterpolatedWind(RPWindTestBase):
    xs = np.linspace(0, 1000, 100) * u.Myr
    lor = LorentzianWind(t0=500 * u.Myr)
    ys = []
    for x in xs:
        ys.append(np.sqrt(np.sum(lor.evaluate(x.value) ** 2, axis=0)))
    ys = np.array(ys)

    wind = InterpolatedWind()
    wind.from_xy(xs, ys)



from .. import events

import galarp as grp


class TestBuiltinUniformEvent:
    
    def test_orbit(self):
        orbits = events.ExampleUniformEvent()

        assert orbits is not None
        assert isinstance(orbits.metadata["WIND"], grp.RPWind)


class TestBuiltinLorentzianEvent:
    
    def test_orbit(self):
        orbits = events.ExampleLorentzianEvent()

        assert orbits is not None
        assert isinstance(orbits.metadata["WIND"], grp.LorentzianWind)



from .general_plots import *      # General plotting functions
from .matrix_plots import *       # Matrix plotting functions
from .utils import *        # Utility functions (colormaps, pyplot styles, etc)



from astropy import units as u


class DensityModel:

    def __init__(self, name=None):
        self.name = name

    def evaluate(self, r):
        raise NotImplementedError


class SphericalBetaModel(DensityModel):
    def __init__(self, n0=0.0121 * u.cm ** -3, r_c=25 * u.kpc, beta=0.655):
        super().__init__("Spherical Beta Model")
        self.n0 = n0
        self.r_c = r_c
        self.beta = beta

    def evaluate(self, r):
        return self.n0 * (1 + (r / self.r_c) ** 2) ** (-3 * self.beta / 2)


class MB2015ModifiedBeta(DensityModel):
    def __init__(self, n0=0.0121 * u.cm ** -3, r_c=25 * u.kpc, beta=0.655):
        super().__init__("Miller and Bregman (2015) Modified Beta Model")
        self.n0 = n0
        self.r_c = r_c
        self.beta = beta

    def evaluate(self, r):
        return self.n0 * self.r_c ** (3 * self.beta) / r ** (3 * self.beta)





class InitConditions:
    """ Container for a satellite's initial conditions."""
    def __init__(self, pos=None, vel=None):
        self.pos = pos
        self.vel = vel

    def __repr__(self):
        return f"InitConditions(pos={self.pos}, vel={self.vel})"

